#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/serial_reg.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "gigaset.h"



/* ----------- basic input functions ------------- */



int Gigaset::readParameter() {

  int err;
  if((err=communicate("&V#C#I0#I1"))<0) return err;
 
  mode=buffer[1][1]-'0';
  rate=buffer[1][4]-'0';
  format=buffer[1][7]-'0';
  handshake=buffer[1][10]-'0';
  type=buffer[1][13]-'0';
  connected=atoi(buffer[3]);
  strength=atoi(buffer[5]);
  quality=atoi(buffer[6]);

  return GIGA_OK;

};



int Gigaset::readStationInfo() {

  int err;
  if((err=readParameter())) return err;
  
  if((err=communicate("I0,0"))<0) return err;
  strcpy(local_hardware,buffer[1]);
  
  if((err=communicate("I1,0"))<0) return err;
  strcpy(local_firmware,buffer[1]);
  
  if((err=communicate("I2,0"))<0) return err;
  strcpy(local_id,buffer[1]);
  
  if((err=communicate("I4,0"))<0) return err;
  strcpy(local_name,buffer[1]);
  
  if(connected) {
    
    if((err=communicate("I0,1"))<0) return err;
    strcpy(partner_hardware,buffer[1]);
    
    if((err=communicate("I1,1"))<0) return err;
    strcpy(partner_firmware,buffer[1]);
    
    if((err=communicate("I2,1"))<0) return err;
    strcpy(partner_id,buffer[1]);
    
    if((err=communicate("I4,1"))<0) return err;
    strcpy(partner_name,buffer[1]);
    
  };
  
  return GIGA_OK;
  
};


  
/* ------------- basic output functions ----------- */



int Gigaset::writeParameter() {

  char cmd[]="B_\\R_\\F_\\Q_";

  if((mode<0) || (mode>4)) return GIGA_ARG_MODE_ERROR;
  cmd[1]=mode+'0';

  if((rate<0) || (rate>9)) return GIGA_ARG_RATE_ERROR;
  cmd[4]=rate+'0';

  if((format<0) || (format>2)) return GIGA_ARG_FORMAT_ERROR;
  cmd[7]=format+'0';

  if((handshake<0) || (handshake>6)) return GIGA_ARG_HANDSHAKE_ERROR;
  cmd[10]=handshake+'0';

  return communicate(cmd);

};



int Gigaset::writeName(char* name,int index=0) {
  
  if(strlen(name)>20) return GIGA_ARG_NAME_ERROR;

  char linebuffer[GIGA_NUM_COLUMNS];
  sprintf(linebuffer,"N%s,%i",name,index);

  return communicate(linebuffer);

};



int Gigaset::writeType() {

  if((type<0) || (type>1)) return GIGA_ARG_TYPE_ERROR;

  char cmd[]="^SETTYPE_";
  cmd[8]=type+'0';

  return communicate(cmd,1);
};



/* --------------- operations ---------------- */



void Gigaset::initSequence() {

  struct {
    int rts_state;
    int dtr_state;
    int delay;
  } phase[] = {
    {0,1,200},   /* Phase 1     */
    {0,0,200},   /* Phase 2     */
    {1,0,100},   /* Phase 3.1.1 */
    {0,0,100},   /* Phase 3.1.2 */
    {1,0,100},   /* Phase 3.2.1 */
    {0,0,100},   /* Phase 3.2.2 */
    {1,0,100},   /* Phase 3.3.1 */
    {0,0,100},   /* Phase 3.3.2 */
    {1,0,100},   /* Phase 3.4.1 */
    {0,0,100},   /* Phase 3.4.2 */
    {1,0,100},   /* Phase 3.5.1 */
    {0,0,100},   /* Phase 3.5.2 */
    {1,1,300},   /* Phase 4     */
    {1,1,500},   /* Phase 5     */
    {0,0,  0}    /* STOP        */
  };

  if(showstat) {printf("Initializing ");fflush(stdout);};

  int i=0;

  while(phase[i].delay!=0) {

    if(showstat) {printf(".");fflush(stdout);};

    int state;
    ioctl(fd,TIOCMGET,&state);
    if(phase[i].rts_state) state|=TIOCM_RTS; else state&=~TIOCM_RTS;
    if(phase[i].dtr_state) state|=TIOCM_DTR; else state&=~TIOCM_DTR;
    ioctl(fd,TIOCMSET,&state);
    usleep(phase[i].delay*1000);

    i++;

  };

  if(showstat) {printf(" done\n");};

};



int Gigaset::startRegistration() {
  return communicate("^ENABREG");
};

int Gigaset::registerStation(char* pin) {

  if(strlen(pin)>5) return GIGA_ARG_PIN_ERROR;
  if(strlen(pin)==0) return GIGA_ARG_PIN_ERROR;

  char linebuffer[GIGA_NUM_COLUMNS];
  sprintf(linebuffer,"^REGISTER%s,0",pin);

  return communicate(linebuffer);

};

int Gigaset::changePIN(char* oldpin,char* newpin) {

  if(strlen(oldpin)>5) return GIGA_ARG_PIN_ERROR;
  if(strlen(oldpin)==0) return GIGA_ARG_PIN_ERROR;
  if(strlen(newpin)>5) return GIGA_ARG_PIN_ERROR;
  if(strlen(newpin)==0) return GIGA_ARG_PIN_ERROR;

  char linebuffer[GIGA_NUM_COLUMNS];
  sprintf(linebuffer,"^CHGPIN%s,%s",oldpin,newpin);

  return communicate(linebuffer);

};



int Gigaset::operate() {
  return communicate("O",1);
};



int Gigaset::reset() {
  return communicate("&F",1);
};



/* ------------ gigaset communication ---------- */



int Gigaset::communicate(char* cmd,int nowait=0) {

  char linebuffer[GIGA_NUM_LINES];

  write(fd,"AT+SDV24",8);
  write(fd,cmd,strlen(cmd));
  write(fd,"\015",1);
  tcdrain(fd);

  if(nowait) return GIGA_OK;

  z=0;s=0;
  while(1) {
    char c;
    read(fd,&c,1);

    if(c==13) continue;
    if(c==10) {

      linebuffer[s]=0;
      s=0;

      strcpy(buffer[z],linebuffer);

      if(strcmp(linebuffer,"OK")==0) return z;
      if(strcmp(linebuffer,"ERROR")==0) return GIGA_ERROR;
      if(strcmp(linebuffer,"FAILURE")==0) return GIGA_FAILURE;
      if(strcmp(linebuffer,"BASE_FULL")==0) return GIGA_BASE_FULL;
      // if(strcmp(linebuffer,"INVALID_PIN")==0) return GIGA_INVALID_PIN;

      if(z<(GIGA_NUM_LINES-1)) z++;

    }
    else {

      // printf("%c\n",c);
      if(s<(GIGA_NUM_COLUMNS-1)) linebuffer[s++]=c;
      
    };

  };

};



/* basics */

Gigaset::Gigaset() :
  mode(0),rate(0),format(0),handshake(0),type(0),
  connected(0),strength(0),quality(0),
  timeout(0),fd(0),device(0),showstat(0)
{

};



Gigaset::~Gigaset() {
  if(fd) closeDevice();
};



int Gigaset::openDevice(char* dev) {

  device=dev;
  if((fd=open(device,O_RDWR /*| O_NOCTTY*/))<0) return GIGA_ERROR;

  bzero(&tio_new,sizeof(struct termios));
  cfmakeraw(&tio_new);
  tio_new.c_cflag = B38400 | CRTSCTS | CS8 | CLOCAL | CREAD;
  tcsetattr(fd,TCSANOW,&tio_new);
  tcflush(fd,TCIOFLUSH);

  setpriority(PRIO_PROCESS,PRIO_PROCESS,-20);
  initSequence();

  tcflush(fd,TCIOFLUSH);

  return GIGA_OK;

};



int Gigaset::closeDevice() {

  // tcsetattr(fd,TCSANOW,&tio_old);
  close(fd);

  return GIGA_OK;

};





